(function() {
  var Server, exports;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  exports = this;
  exports.Server = Server = (function() {
    function Server(args) {
      var io;
      this.args = args != null ? args : {};
      this.queue_entity_state_change = __bind(this.queue_entity_state_change, this);
      this.on_game_tick_complete = __bind(this.on_game_tick_complete, this);
      io = this.args.io;
      this.sockets_by_id = {};
      this.queued_messages_for_socket_id = {};
      this.changed_entity_states_by_entity_id = {};
      this.game = this.args.game;
      this.game.on('entity_created', __bind(function(entity) {
        return _.extend(entity, new Lerpable());
      }, this));
      this.game.on('entity_changed', __bind(function(changed_state) {
        return this.queue_entity_state_change(changed_state);
      }, this));
      this.game.on('tick', __bind(function(args) {
        return this.on_game_tick_complete({
          gametime: args.gametime
        });
      }, this));
      io.sockets.on('connection', __bind(function(socket) {
        return this.on_connect(socket);
      }, this));
    }
    Server.prototype.on_connect = function(socket) {
      var player;
      player = this.game.add_player(socket.id);
      socket.player = player;
      player.socket = socket;
      this.sockets_by_id[socket.id] = socket;
      this.send_init_game_info_to_socket(socket);
      this.send_all_entities_to_socket(socket);
      this.send_new_player_to_other_sockets(player);
      socket.on('message', __bind(function(message) {
        return this.on_message(socket, message);
      }, this));
      return socket.on('disconnect', __bind(function() {
        return this.on_disconnect(socket);
      }, this));
    };
    Server.prototype.on_message = function(socket, message) {
      var msg;
      this.log("IN: " + message);
      msg = JSON.parse(message);
      return this.handle_message(socket, msg);
    };
    Server.prototype.on_disconnect = function(socket) {
      var msg;
      this.game.remove_player(socket.player.id);
      delete this.sockets_by_id[socket.id];
      delete this.queued_messages_for_socket_id[socket.id];
      delete this.changed_entity_states_by_entity_id[socket.player.id];
      msg = new Message({
        message_type: Message.REMOVE_ENTITY,
        entity_id: socket.player.id
      });
      return this.send_message_to_all_except_socket(msg, socket);
    };
    Server.prototype.handle_message = function(socket, msg) {
      var state;
      switch (msg.message_type) {
        case Message.MESSAGE_TYPES.ACK:
          return this.on_ack(socket, msg);
        case Message.MESSAGE_TYPES.PLAYER_COMMAND:
          socket.player[msg.command](msg.args);
          state = socket.player.get_state();
          msg = new Message({
            message_type: Message.MESSAGE_TYPES.ACK,
            command_id: msg.command_id,
            recieved_at: new Date().getTime(),
            entity_state: state
          });
          return this.queue_message_for_socket(msg, socket);
      }
    };
    Server.prototype.queue_message_for_socket = function(msg, socket) {
      var _base, _name;
      return ((_base = this.queued_messages_for_socket_id)[_name = socket.id] || (_base[_name] = [])).push(msg);
    };
    Server.prototype.send_message_to_socket = function(msg, socket) {
      if (!socket) {
        return;
      }
      msg = JSON.stringify(msg);
      socket.send(msg);
      return this.log("OUT: " + msg);
    };
    Server.prototype.send_all_entities_to_socket = function(socket) {
      var entity, entity_id, entity_states_by_entity_id, msg, _ref;
      entity_states_by_entity_id = {};
      _ref = this.game.entities_by_id;
      for (entity_id in _ref) {
        entity = _ref[entity_id];
        entity_states_by_entity_id[entity_id] = entity.get_state();
      }
      msg = new Message({
        message_type: Message.MESSAGE_TYPES.UPDATE_ENTITIES,
        at_gametime: this.game.gametime,
        entity_states_by_entity_id: entity_states_by_entity_id
      });
      return this.queue_message_for_socket(msg, socket);
    };
    Server.prototype.send_init_game_info_to_socket = function(socket) {
      var msg;
      msg = new Message({
        at_gametime: this.game.gametime,
        message_type: Message.MESSAGE_TYPES.GAME_INIT,
        your_entity_id: socket.player.id,
        your_entity_state: socket.player.get_state()
      });
      return this.queue_message_for_socket(msg, socket);
    };
    Server.prototype.send_new_player_to_other_sockets = function(player) {
      var changed_state, id, socket, _ref, _results;
      _ref = this.sockets_by_id;
      _results = [];
      for (id in _ref) {
        socket = _ref[id];
        if (socket !== player.socket) {
          changed_state = {
            entity_id: player.id,
            entity_state: player.get_state()
          };
          _results.push(this.queue_entity_state_change(changed_state));
        }
      }
      return _results;
    };
    Server.prototype.send_message_to_all_except_socket = function(msg, except_socket) {
      var id, socket, _ref, _results;
      _ref = this.sockets_by_id;
      _results = [];
      for (id in _ref) {
        socket = _ref[id];
        if (socket !== except_socket) {
          _results.push(this.send_message_to_socket(msg, socket));
        }
      }
      return _results;
    };
    Server.prototype.on_game_tick_complete = function(args) {
      var id, message, messages_to_send, msg, queued_messages, socket, socket_id, _i, _len, _ref, _ref2, _ref3, _results;
      _ref = this.queued_messages_for_socket_id;
      for (socket_id in _ref) {
        queued_messages = _ref[socket_id];
        socket = this.sockets_by_id[socket_id];
        _ref2 = [queued_messages, []], messages_to_send = _ref2[0], this.queued_messages_for_socket_id[socket_id] = _ref2[1];
        for (_i = 0, _len = messages_to_send.length; _i < _len; _i++) {
          message = messages_to_send[_i];
          this.send_message_to_socket(message, socket);
        }
      }
      if (this.changed_entity_states) {
        msg = new Message({
          message_type: Message.MESSAGE_TYPES.UPDATE_ENTITIES,
          at_gametime: args.gametime,
          entity_states_by_entity_id: this.changed_entity_states_by_entity_id
        });
        this.changed_entity_states_by_entity_id = {};
        this.changed_entity_states = false;
        _ref3 = this.sockets_by_id;
        _results = [];
        for (id in _ref3) {
          socket = _ref3[id];
          _results.push(this.send_message_to_socket(msg, socket));
        }
        return _results;
      }
    };
    Server.prototype.queue_entity_state_change = function(changed_entity_state) {
      var changed_prop, changed_value, merged_entity_state, _base, _name, _ref;
      merged_entity_state = ((_base = this.changed_entity_states_by_entity_id)[_name = changed_entity_state.entity_id] || (_base[_name] = {
        entity_id: changed_entity_state.entity_id
      }));
      _ref = changed_entity_state.entity_state;
      for (changed_prop in _ref) {
        changed_value = _ref[changed_prop];
        merged_entity_state[changed_prop] = changed_value;
      }
      return this.changed_entity_states = true;
    };
    Server.prototype.log = function(data) {
      return sys.log(data);
    };
    return Server;
  })();
}).call(this);
