#= require './message'
#= require './lerpable'
#= require '../vendor/underscore'

sys = require 'sys'
_ = require '../vendor/underscore.min'
percept = require './percept'

exports = this

exports.Server = class Server
  constructor: (@args={}) ->
    io = @args.io

    @sockets_by_id = {}

    #Whenever a game object notifies us about its changes, we'll remember the object and it's changed values so we can broadcast its changed states on the next tick
    @queued_messages_for_socket_id = {}
    @changed_entity_states_by_entity_id = {}

    #Handle events that the game will emit to us for serving
    @game = @args.game
    @game.on 'entity_created', (entity) => _.extend(entity, new percept.Lerpable())
    @game.on 'entity_changed', (changed_state) => @queue_entity_state_change(changed_state)
    @game.on 'tick', (args) => @on_game_tick_complete(gametime: args.gametime)

    #Handle connections
    io.sockets.on 'connection', (socket) =>
      @on_connect(socket)


  on_connect: (socket) ->
    player = @game.add_player(socket.id)
    socket.player = player
    player.socket = socket
    @sockets_by_id[socket.id] = socket

    @send_init_game_info_to_socket(socket)
    @send_all_entities_to_socket(socket)
    @send_new_player_to_other_sockets(player)

    socket.on 'message', (message) =>
      @on_message(socket, message)

    socket.on 'disconnect', =>
      @on_disconnect(socket)


  on_message: (socket, message) ->
    @log "IN: #{message}"
    msg = JSON.parse(message)
    @handle_message(socket, msg)


  on_disconnect: (socket) ->
    @game.remove_player(socket.player.id)
    delete @sockets_by_id[socket.id]
    delete @queued_messages_for_socket_id[socket.id]
    delete @changed_entity_states_by_entity_id[socket.player.id]

    msg = new percept.Message
      message_type: percept.Message.REMOVE_ENTITY
      entity_id: socket.player.id
    @send_message_to_all_except_socket(msg, socket)


  handle_message: (socket, msg) ->
    switch msg.message_type
      # TODO: decide if we care about generic ACK msgs from clients
      when percept.Message.MESSAGE_TYPES.ACK
        return @on_ack(socket, msg)

      when percept.Message.MESSAGE_TYPES.PLAYER_COMMAND
        # Invoke the messaged command onto the player object, passing in messaged args
        socket.player[msg.command](msg.args)

        # Acknowledge this player command back to the client along with the player's new state
        #
        # THIS WONT WORK, SEE BELOW FOR WHY: 
        # changed_state = @changed_entity_states_by_entity_id[socket.player.id]
        #
        # TODO: Figure out how to only send back the changed state of the player.  
        #       We can't rely on @changed_entity_states_by_entity_id above because player's changed callback
        #       hasn't had a chance to replicate back into it yet.  So for now we'll
        #       just return the player's entire state which is not network efficient.
        state = socket.player.get_state()
        msg = new percept.Message
          message_type: percept.Message.MESSAGE_TYPES.ACK
          command_id: msg.command_id
          recieved_at: new Date().getTime()
          entity_state: state
        @queue_message_for_socket msg, socket


  queue_message_for_socket: (msg, socket) ->
    (@queued_messages_for_socket_id[socket.id] ||= []).push msg


  send_message_to_socket: (msg, socket) ->
    return unless socket
    msg = JSON.stringify(msg)
    socket.send msg
    @log "OUT: #{msg}"


  send_all_entities_to_socket: (socket) ->
    entity_states_by_entity_id = {}

    for entity_id, entity of @game.entities_by_id
      entity_states_by_entity_id[entity_id] = entity.get_state()

    msg = new percept.Message
      message_type: percept.Message.MESSAGE_TYPES.UPDATE_ENTITIES
      at_gametime: @game.gametime
      entity_states_by_entity_id: entity_states_by_entity_id
    @queue_message_for_socket(msg, socket)


  send_init_game_info_to_socket: (socket) ->
    msg = new percept.Message
      at_gametime: @game.gametime
      message_type: percept.Message.MESSAGE_TYPES.GAME_INIT
      your_entity_id: socket.player.id
      your_entity_state: socket.player.get_state()
    @queue_message_for_socket msg, socket


  send_new_player_to_other_sockets: (player) ->
    for id, socket of @sockets_by_id when socket != player.socket
      #TODO refactor out this code duplication
      changed_state = 
        entity_id: player.id
        entity_state: player.get_state()
      @queue_entity_state_change changed_state


  send_message_to_all_except_socket: (msg, except_socket) ->
    @send_message_to_socket(msg, socket) for id, socket of @sockets_by_id when socket != except_socket


  on_game_tick_complete: (args) =>
    # Send each socket all of its queued messages
    for socket_id, queued_messages of @queued_messages_for_socket_id
      #@log "Queued messages for socket_id #{socket_id}: #{queued_messages}"
      socket = @sockets_by_id[socket_id]
      [messages_to_send, @queued_messages_for_socket_id[socket_id]] = [queued_messages, []]
      @send_message_to_socket(message, socket) for message in messages_to_send

    #Also send the UPDATE_ENTITIES message to all sockets
    if @changed_entity_states
      msg = new percept.Message
        message_type: percept.Message.MESSAGE_TYPES.UPDATE_ENTITIES
        at_gametime: args.gametime
        entity_states_by_entity_id: @changed_entity_states_by_entity_id
      @changed_entity_states_by_entity_id = {}
      @changed_entity_states = false
      @send_message_to_socket(msg, socket) for id, socket of @sockets_by_id


  queue_entity_state_change: (changed_entity_state) =>
    #store the latest version of the changed properties of this entity for sharing on next tick
    merged_entity_state = (@changed_entity_states_by_entity_id[changed_entity_state.entity_id] ||= {entity_id: changed_entity_state.entity_id})
    merged_entity_state[changed_prop] = changed_value for changed_prop, changed_value of changed_entity_state.entity_state
    @changed_entity_states = true


  log: (data) ->
    sys.log data

